// Copyright: Dale Schinstock, 2013

#include "GloOS.h"

// *******************************
// GloOS Class
// *******************************

GloOS::GloOS(uint16_t sysTimerFrequency,
          void (*sysTimerThread)(void),
          void (*thread1Ptr)(void),
          void (*thread2Ptr)(void)): telemQueue(GLOOSTELEMQUEUESIZE)
{
  float thread2frequency = 1.0f;
  
  threadFuncPtrs[0] = sysTimerThread;
  threadFuncPtrs[1] = thread1Ptr;
  threadFuncPtrs[2] = thread2Ptr;
  
  __disable_irq();  // disable interrupts
  
  EXTI_InitTypeDef EXTI_InitStructure;
  NVIC_InitTypeDef NVIC_InitStructure;

  sysTimerSetup(sysTimerFrequency); // Sets up interrupt
  thread2delay = (uint32_t)(sysTimer.freq/thread2frequency); //TODO - check bounds
    
  // Usually connect a pin to EXTI, but we are using as software interrupt
  // Configure EXTI lines 
  EXTI_InitStructure.EXTI_Line = EXTI_Line5;
  EXTI_InitStructure.EXTI_Mode = EXTI_Mode_Interrupt;
  EXTI_InitStructure.EXTI_Trigger = EXTI_Trigger_Rising; // shouldn't matter for SW intpt  
  EXTI_InitStructure.EXTI_LineCmd = ENABLE;
  EXTI_Init(&EXTI_InitStructure);
  EXTI_InitStructure.EXTI_Line = EXTI_Line10;
  EXTI_Init(&EXTI_InitStructure);
  

  // Enable EXTI Interrupts
  NVIC_InitStructure.NVIC_IRQChannel = EXTI9_5_IRQn;
  NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 0x07;
  NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0x00;   // subpriority not used for us? 
  NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
  NVIC_Init(&NVIC_InitStructure); 
  NVIC_InitStructure.NVIC_IRQChannel = EXTI15_10_IRQn;
  NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 0x08;
  NVIC_Init(&NVIC_InitStructure); 

  // Use all bits of interrupt priority to be preemption (16 priority levels)
  SCB->AIRCR = 0x05FA0300;   

  // THREAD INTERRUPTS - systick should be highest priority
  SCB->SHP[11]=0x20;              // Make Systick third highest priority interrupt
  NVIC->IP[EXTI9_5_IRQn] = 0x70;  // Make EXTI5 thread 8th highest priority
  NVIC->IP[EXTI15_10_IRQn] = 0x80;// Make EXTI10 thread 9th highest priority
  
  // NONTHREAD INTERRUPTS
  NVIC->IP[OTG_FS_IRQn] = 0x60;   // Put USB (OTG_FS) interrupt at 7th highest priority   
                                  // SHOULD USB BE HIGHER PRIORITY? - check timing on scope
  
  __enable_irq();  // enable interrupts
}

void GloOS::threadDispatcher(threadEnum thread)      // reentrant function
{
  switch(thread)
  {
    case thread0:
      leds.set(orange);
      sysTimer.ticks++;
      sysTimer.time = (double)sysTimer.ticks/sysTimer.freq;
      
      if (threadFuncPtrs[0] != 0)
        threadFuncPtrs[0]();

      threadScheduler();
      leds.clear(orange);
      break;

    case thread1:
      leds.set(blue);
      threadStats.thread1Count++;
      if (threadFuncPtrs[1] != 0) 
        threadFuncPtrs[1]();
      leds.clear(blue);
      break;
      
    case thread2:
//      leds.set(red);
      threadStats.thread2Count++;
      if (threadFuncPtrs[2] != 0) 
        threadFuncPtrs[2]();
//      leds.clear(red);
      break;
  }
}  

void GloOS::threadScheduler(void)      // not reentrant  
{                                      // called only by sysTick
  static uint32_t thread2counter;

  if (telemQueue.getnum())                  // redundant with telemQueue.put()
    EXTI->SWIER |= EXTI_Line5;              // to prevent lockouts
  
  thread2counter++;
  if (thread2counter >= thread2delay)
  {
    thread2counter = 0;
    EXTI->SWIER |= EXTI_Line10;              // generate interrupt
  }
}

void GloOS::delay(float seconds) // reentrant and threadsafe (not sysTimerThread)
{
  uint64_t start;

  start = sysTimer.ticks;
  if (seconds < 1.5f*sysTimer.DT)
  {
   while (sysTimer.ticks == start);
  }
  else
    while( ((float)(sysTimer.ticks-start)/sysTimer.freq) < seconds);
}

void GloOS::sysTimerSetup(uint16_t SysTickFrequency)
{
  RCC_ClocksTypeDef RCC_Clocks;

  sysTimer.time = 0;
  sysTimer.ticks = 0;
  sysTimer.freq = SysTickFrequency;
  sysTimer.DT = 1.0f/sysTimer.freq;

  // Use all bits of interrupt priority to be preemption (16 priority levels)
  SCB->AIRCR = 0x05FA0300;   

  /* SysTick configuration */
  RCC_GetClocksFreq(&RCC_Clocks);
  SysTick_Config(RCC_Clocks.HCLK_Frequency / SysTickFrequency);
  SCB->SHP[11]=0x20;     // Make Systick third highest priority interrupt
}

uint8_t GloOS::mutexLock(uint8_t mutexNum)  // TODO: should implement ownership
{
   __disable_irq();
   if ((mutexNum<GLOOSNUMMUTEX) && !(mutex[mutexNum]))
   {
     mutex[mutexNum]=1;
     __enable_irq();
     return(1);
   }
   else
   {
      __enable_irq();
      return(0);
   }
}

// *******************************
// GloObjectQueue Class
// *******************************

GloObjectQueue::GloObjectQueue(uint8_t thissize)
{
  queue = new gloObjectQueueType[thissize];
  num=0;
  top=0;
  bottom=0;
  size = thissize;
}

uint8_t GloObjectQueue::put(uint8_t objID, uint16_t instance)  // reentrant and threadsafe
{
  if ((objID <= NUMGLOTYPES)&&(instance <= gloObjectPtrs[objID]->md.numInstance))
  {
     __disable_irq();
     num++;
     if (num > size)
     {
       num--;
       __enable_irq();
       return(0);
     }
     top++;
     if (top == size) top = 0;
     queue[top].objID = objID;
     queue[top].instance = instance;

     EXTI->SWIER |= EXTI_Line5;    // Generate EXTI SW interrupt (run thread)
     __enable_irq();
     return(1);
  }
  else return(0);
}

uint8_t GloObjectQueue::take(uint8_t *objID, uint16_t *instance) // reentrant and threadsafe
{
  if(num)
  {
    __disable_irq();
    bottom++;
    if (bottom == size) bottom = 0;
    num--;
    *objID = queue[bottom].objID;
    *instance = queue[bottom].instance;
    __enable_irq();
    return(1);
  }
  else return(0);
}
