// Motorlab Control Program for STM32F4-Discovery
// Copyright: Dale Schinstock, 2013

#include "GloLink.h"
#include <math.h>
#include "stm32f4xx.h"
#include "sysTimerThread.h"
#include <stddef.h>
#include <string.h>
#include "GloOS.h"
#include "telemSendThread.h"

//************************************************************************
//************************************************************************
#define sysTimerFrequency 10000

//************************************************************************
//**************   internal function prototypes **************************

void thread2Thread(void);
void trapProfReceiveCallback(uint16_t instance, void *data);
void parametersReceiveCallback(uint16_t instance, void *data);
void statusReceiveCallback(uint16_t instance, void *data);
void initState(void);


//************************************************************************
//**************** GLOBAL VARIABLES **************************************

//************************************************************************
//**************** GLOBAL OBJECTS DECLARED EXTERN IN THEIR HEADERS  ******
GloLink gloLink;
GloOS gloOS(sysTimerFrequency, sysTimerThread,telemSendThread,thread2Thread);

//************************************************************************
//************************************************************************
int main(){
  
  initState();
  gloRegisterOnReceiveFunc(trapProf.md.id, trapProfReceiveCallback);
  gloRegisterOnReceiveFunc(parameters.md.id, parametersReceiveCallback);
  gloRegisterOnReceiveFunc(status.md.id, statusReceiveCallback);

  while(1)
  {
    gloLink.checkForObjects();  // the idle task just handles recieved messages
  }
}

//************************************************************************
//************************************************************************
void thread2Thread(void){
  static uint32_t sendCounts;
  
//  gloOS.delay(0.5);  // this is bad, blocks all lower priority threads

  gloOS.leds.toggle(green);
  
  if (sendCounts == gloOS.threadStats.thread1Count)
    gloOS.leds.set(red);
  else
    gloOS.leds.clear(red);
  
  sendCounts = gloOS.threadStats.thread1Count;
}

//************************************************************************
//************************************************************************
void trapProfReceiveCallback(uint16_t instance, void *data)
{
  trapProfDataType *sent = (trapProfDataType *)data;
  float deltaS,deltaT1,deltaT2;

   deltaS = (sent->Vmax*sent->Vmax)/sent->Amax;  // the standard to decide if the deltaT2=0
   if (fabsf(sent->DX) < deltaS){
     deltaT1 = sqrtf(fabsf(sent->DX)/sent->Amax);// find the time splits for the three
     deltaT2 = 0;                                // parabolic splines
     sent->t1 = deltaT1;
     sent->t3 = sent->t1+deltaT1;
     sent->t2 = 0;                               // DON'T USE MIDDLE SPLINE
     sent->c1[0] = sent->Amax/2.0f;              // coefficients of the first parobolic spline
     sent->c1[1] = 0;
     sent->c1[2] = 0;          
     sent->c3[0] = -sent->Amax/2.0f;             // coefficients of the 3rd parobolic spline
     sent->c3[1] = sent->Amax*sent->t3;   
     sent->c3[2] = sent->DX-1.0f/2.0f*sent->Amax*(sent->t3*sent->t3);
  }
  else{
    deltaT1 = sent->Vmax/sent->Amax;             // find the time splits for the three
    deltaT2 = fabsf(sent->DX)/sent->Vmax-deltaT1;// parabolic splines
    sent->t1 = deltaT1;
    sent->t2 = sent->t1 + deltaT2;
    sent->t3 = sent->t2 + deltaT1;
    sent->c1[0] = sent->Amax/2.0f;              // coefficients of the first parobolic spline
    sent->c1[1] = 0;
    sent->c1[2] = 0;           
    sent->c2[0] = 0;                            // coefficients of the 2nd parobolic spline
    sent->c2[1]=sent->Amax*(2.0f*sent->t1*sent->t2-sent->t3*sent->t1)/(sent->t2-sent->t1);
    sent->c2[2]=(sent->t1*sent->t1)*(sent->c1[0]-sent->c2[0])+sent->t1*(sent->c1[1]-sent->c2[1])+sent->c1[2];
    sent->c3[0] = -sent->Amax/2.0f;             // coefficients of the 3rd parobolic spline
    sent->c3[1] = sent->Amax*sent->t3; 
    sent->c3[2] = sent->DX-1.0f/2.0f*sent->Amax*(sent->t3*sent->t3);
  }
  
  trapProf.setData(sent);
}

//************************************************************************
//************************************************************************
void statusReceiveCallback(uint16_t instance, void *data)
{
  statusDataType *received = (statusDataType *)data;
  
  //ONLY REASON FOR GUI TO SEND STATUS IS TO SET THE COMMAND
  status.data.command = received->command;
 }
//************************************************************************
//************************************************************************
void parametersReceiveCallback(uint16_t instance, void *data)
{
  parametersDataType *received = (parametersDataType *)data;
  parametersDataType paramsd;
  
  parameters.getData(&paramsd);
  
  switch (received->guiCmd)
  {
  case GAINS:
    paramsd.Kp=received->Kp; paramsd.Ki=received->Ki; paramsd.Kd=received->Kd;
    break;
  case WMAG:
    paramsd.waveMag=received->waveMag;
    break;
  case WFREQ:
    if (received->waveFreq > gloOS.sysTimer.freq/10.0f)
      paramsd.waveFreq = gloOS.sysTimer.freq/10.0f;
    else if (received->waveFreq < 0.01)
      paramsd.waveFreq = 0.01;
    else
      paramsd.waveFreq = received->waveFreq;
    break;
  case WTYPE:
    paramsd.waveType = received->waveType;
    if (paramsd.waveType==TRAPEZOIDAL){
      // set an approximate waveFreq
      float period = trapProf.data.t3 + trapProf.data.t1;
      if (period < (gloOS.sysTimer.DT*10.0f))
        paramsd.waveFreq = gloOS.sysTimer.freq/10.0f; 
      else
        paramsd.waveFreq = 1.0f/period;
      // try to choose a reasonable sample freq
      float delay = 1.5f*period/logData.md.numInstance/gloOS.sysTimer.DT;
      if (delay < 1.0f)
        paramsd.sampleDelay = 1;
      else
        paramsd.sampleDelay = (uint16_t)delay;
      paramsd.sampleFreq = gloOS.sysTimer.freq/paramsd.sampleDelay; 
    }
    break;
  case CMODE:
    gloOS.flags[GLOOSRESETINTFLAG] = 1;
    paramsd.controlMode = received->controlMode;
    paramsd.enableAmp = false;
    if((paramsd.controlMode != POSITION)&&(paramsd.waveType==TRAPEZOIDAL))
      paramsd.waveType=SQUARE;
    break;
  case FSENSOR:
    paramsd.feedbackSensor = received->feedbackSensor;
    break;
  case AMP:
    gloOS.flags[GLOOSRESETINTFLAG] = 1;
    paramsd.enableAmp = received->enableAmp;
    break;
  case SFREQ:
    if (received->sampleFreq > gloOS.sysTimer.freq){
      paramsd.sampleDelay = 1;
    }
    else if (received->sampleFreq < 10.0f){
      paramsd.sampleDelay = (uint16_t)(gloOS.sysTimer.freq/10.0f);
    }
    else{
      paramsd.sampleDelay = (uint16_t)(gloOS.sysTimer.freq/received->sampleFreq + 0.5);
    }
    paramsd.sampleFreq = gloOS.sysTimer.freq/paramsd.sampleDelay; 
    break;
  case STARTWAVE:
    __disable_irq();
    if (status.data.waveState==STOPPED){
      status.data.commandOffset = status.data.command;
      status.data.waveState = CONTINUOUS;
      status.data.waveTime = 0.0f;
    }
    __enable_irq();
    break;
  case STOPWAVE:
    __disable_irq();
    if (status.data.waveState==CONTINUOUS){
      status.data.commandOffset = status.data.command;
      status.data.waveState = STOPPED;
    }
    break;
    __enable_irq();
  case STARTAUTO:
    __disable_irq();
    if (status.data.waveState==STOPPED){
      status.data.commandOffset = status.data.command;
      status.data.waveState = AUTOSTART;
    }
    __enable_irq();
    break;
  case RESETENC:
    gloOS.flags[GLOOSRESETENCFLAG] = true;
    break;
  }
 
  parameters.setData(&paramsd);
}
//************************************************************************
//************************************************************************
void initState(void)
{
 trapProfDataType tpData;
  
  tpData.DX = 12000;
  tpData.Amax = 6000;
  tpData.Vmax = 6000;
  trapProfReceiveCallback(1,&tpData);

  status.data.command =0.0f;
  status.data.commandOffset=0.0f;
  status.data.waveTime=0.0f;
  status.data.logDataPtr=0.0f;
  status.data.waveState=STOPPED;

  parameters.data.Kp=0.01f;
  parameters.data.Ki=0.0f;
  parameters.data.Kd=0.0f;
  parameters.data.waveMag=1.0f;
  parameters.data.waveFreq=1.0f;
  parameters.data.waveType=SQUARE;
  parameters.data.controlMode=POSITION;
  parameters.data.feedbackSensor=MOTOR;
  parameters.data.enableAmp=false;
  parameters.data.sampleDelay = (uint16_t)(gloOS.sysTimer.freq/500.0f + 0.5);
  parameters.data.sampleFreq = gloOS.sysTimer.freq/parameters.data.sampleDelay;
}