// Copyright: Dale Schinstock, 2013

#include "sysTimerThread.h"
#include "AnalogInPB0.h"
#include "Encoder.h"
#include "DigitalOut.h"
#include "PwmAnalogOut.h"
#include "DerivativeFilter.h"
#include "trigtables.h"
#include "GloLink.h"
#include <math.h>
#include "stm32f4xx.h"
#include "GloOS.h"

AnalogInPB0 currentSense;
Encoder motorEnc(Encoder1);
Encoder loadEnc(Encoder2);
DigitalOut ampEnable(PB11,GPIO_OType_OD,GPIO_PuPd_NOPULL,1);
PwmAnalogOut currentCommand;
DerivativeFilter motorDeriv(gloOS.sysTimer.DT , 47.7465f, 0.3535f);  // IS GLOOS PROPERTY VALID AT THIS POINT?
DerivativeFilter loadDeriv(gloOS.sysTimer.DT , 47.7465f, 0.3535f);
float pidIntegral = 0.0f;  // integral part of controller - global for reset

void storeData(statusDataType* statusd, float U, float I, uint16_t delay);
void genCommand(statusDataType* statusd,
    parametersDataType* paramsd,
    trapProfDataType* trapProfd);
float pid(float Kp, float Ki, float Kd,
         float error, float *integral,
         float DT, float Umax);     // pid controller returns integral outputs U

float motCurrent, currentOut;

//*****************************************************************************
void sysTimerThread(void)
{
 parametersDataType paramsd;
 statusDataType statusd;
 trapProfDataType trapProfd;
  
 float error;
 const float Umax = 3.0f;
  
  parameters.getData(&paramsd);
  status.getData(&statusd);
  trapProf.getData(&trapProfd);
  
  if(gloOS.flags[GLOOSRESETENCFLAG])
  {
    motorEnc.set(0); loadEnc.set(0);
    statusd.command = 0;
    gloOS.flags[GLOOSRESETENCFLAG]=false;
  }

  ampEnable.write(!paramsd.enableAmp);  // turn amp on/off
  if ((gloOS.flags[GLOOSRESETINTFLAG])||(!paramsd.enableAmp)){
    pidIntegral = 0.0f;
    gloOS.flags[GLOOSRESETINTFLAG] = 0;
  }
  motCurrent = -17.402*(currentSense.getAvgVoltage()) + 30.4211;
  statusd.motPos = motorEnc.read()*0.225f;
  statusd.loadPos = loadEnc.read()*0.18f;
  statusd.motSpeed = motorDeriv.velocity(statusd.motPos)/6.0f; // 1rpm/(6deg/s)
  statusd.loadSpeed = loadDeriv.velocity(statusd.loadPos)/6.0f;

  genCommand(&statusd, &paramsd, &trapProfd);

  switch (paramsd.controlMode)
  {
    case OPENLOOP:
      currentOut = statusd.command;
      break;
    case POSITION:
      if (paramsd.feedbackSensor == MOTOR)
        error = statusd.command - statusd.motPos;
      else
        error = statusd.command - statusd.loadPos;
        currentOut = pid(paramsd.Kp, paramsd.Ki, paramsd.Kd,
           error, &pidIntegral,
           gloOS.sysTimer.DT, Umax);
      break;
    case SPEED:
      if (paramsd.feedbackSensor == MOTOR)
        error = statusd.command - statusd.motSpeed;
      else
        error = statusd.command - statusd.loadSpeed;
        currentOut = pid(paramsd.Kp, paramsd.Ki, paramsd.Kd,
           error, &pidIntegral,
           gloOS.sysTimer.DT, Umax);
      break;
  }
  
  currentCommand.setVoltage(currentOut);
  
  if (statusd.waveState != AUTOPAUSE)
    if (gloOS.mutexLock(GLOOSDATAMUTEX))
    {
      storeData(&statusd, currentOut, motCurrent, paramsd.sampleDelay);
      gloOS.mutexUnlock(GLOOSDATAMUTEX);
    }

  status.setData(&statusd);  
  parameters.setData(&paramsd);
}

//*****************************************************************************
void genCommand(statusDataType* statusd,
    parametersDataType* paramsd,
    trapProfDataType* trapProfd)
{
 float *command,*waveTime,period,waveMag,comOffset,waveFreq,junk;
  
  command = &(statusd->command);    // this is changed in this function
  waveTime = &(statusd->waveTime);   // this is changed in this function

  waveMag = paramsd->waveMag;
  comOffset = statusd->commandOffset;
  waveFreq = paramsd->waveFreq;
  period = 1.0f/waveFreq;
  
  *waveTime += gloOS.sysTimer.DT;
  if ((*waveTime >= period)&&(paramsd->waveType != TRAPEZOIDAL))
    *waveTime -= period;
  if (statusd->waveState == AUTOSTART) *waveTime = -period/8.0f;
  
  if ((statusd->waveState == AUTOSAVE)&&(*waveTime < 0.0f)){
    *command = comOffset;
    return;
  }
  
  if ((statusd->waveState == AUTOSAVE)||(statusd->waveState == CONTINUOUS))
    switch (paramsd->waveType)
    {
    case SINE:
      sincostab(*waveTime*waveFreq*6.283185, command, &junk);
      *command = comOffset + waveMag*(*command);
      break;
    case SQUARE:
      if (*waveTime < 0.0f) *command = comOffset;
      else if (*waveTime < 0.5f*period) *command = comOffset + waveMag;
      else *command = comOffset;
      break;
    case TRIANGLE:
      if (*waveTime < 0.25*period)
        *command = comOffset + 4.0f*waveMag*waveFreq*(*waveTime);
      else if (*waveTime < 0.75*period)
        *command = comOffset + 2.0f*waveMag*(1.0f-2.0f*waveFreq*(*waveTime));
      else
        *command = comOffset + 4.0f*waveMag*(waveFreq*(*waveTime)-1.0f);
      break;
    case TRAPEZOIDAL:
      if (*waveTime <0)
        *command = comOffset;
      else if (*waveTime <= trapProfd->t1){                            // the first spline function
        *command = comOffset + trapProfd->c1[0]*(*waveTime**waveTime) +
        trapProfd->c1[1]**waveTime + trapProfd->c1[2];
      }
      else if((*waveTime>trapProfd->t1)&&(*waveTime<=trapProfd->t2)){   // the second spline function
        *command = comOffset + trapProfd->c2[0]*(*waveTime**waveTime) +
        trapProfd->c2[1]**waveTime + trapProfd->c2[2];
      }
      else if((*waveTime>trapProfd->t2)&&(*waveTime<=trapProfd->t3)){   // the third spline function
        *command = comOffset + trapProfd->c3[0]*(*waveTime**waveTime) +
        trapProfd->c3[1]**waveTime + trapProfd->c3[2];
      }
      if (*waveTime > trapProfd->t3){
        *command = comOffset + trapProfd->DX;
      }
      break;
    }
}

//*****************************************************************************
void storeData(statusDataType* statusd, float U, float I, uint16_t delay)
{
 static uint16_t count=0;
 static float dataTime=0, waveZeroMark=0;
 logDataDataType *data;
 float dT = delay*gloOS.sysTimer.DT;  // delta time for data storage
  
  data = &(logData.data[0]);    // we will write directly to logData glo object
 
  count++;
  if (count >= delay){
    count = 0;
    statusd->logDataPtr++; //logDataPtr should always point to last vals written
    waveZeroMark = 0.0f;
  } 

  if (statusd->waveState == AUTOSTART){    //start at the beginning of buffer
    count=0;
    statusd->logDataPtr = 0;
    statusd->waveState = AUTOSAVE;
    waveZeroMark = 0.0f;
  }

  if (statusd->logDataPtr >= (NUMLOGDATA)){ 
    if (statusd->waveState == AUTOSAVE){    // stop saving and retract pointer
      statusd->waveState = AUTOPAUSE;
      statusd->logDataPtr--;
      return;
    }
    statusd->logDataPtr = 0; 
  }
  dataTime = statusd->logDataPtr*dT;
  
  if ((statusd->waveTime <= dT/2.0f)&&(statusd->waveTime > -dT/2.0f))
    waveZeroMark = 1.0f;              // mark the points where waveTime is zero
  
  uint16_t ptr=statusd->logDataPtr;
  data[ptr].floatvals[0]=dataTime;
  data[ptr].floatvals[1]=statusd->command;
  data[ptr].floatvals[2]=statusd->motPos;
  data[ptr].floatvals[3]=statusd->loadPos;
  data[ptr].floatvals[4]=statusd->motSpeed;
  data[ptr].floatvals[5]=statusd->loadSpeed;
  data[ptr].floatvals[6]=U;
  data[ptr].floatvals[7]=I;
  data[ptr].floatvals[8]=waveZeroMark;
}

//*****************************************************************************
float pid(float Kp, float Ki, float Kd,
         float error, float *integral,
         float DT, float Umax)      // pid controller returns integral outputs U
{
 static float lastError;
 float U, USat;
 
  *integral = *integral + Ki*DT*error;     // Integral portion of PID
  USat = Umax;
  if (*integral > USat) *integral = USat;    // stopgap limits on integral
  if (*integral < -USat) *integral = -USat;  // to limit integral windup 

  U = *integral + Kp*error + Kd*(error-lastError)/DT;   // PID 
  USat = Umax;
  if (U > USat) U = USat;               // saturation limits
  if (U < -USat) U = -USat;

  lastError = error;      // remember last error
  return(U);
}                                               